==> v0.8.1.zip: shellfu/src/include/pretty.sh <==
#!/bin/bash

shellfu import exit

#
# Name of pretty-printer module
#
# Friendly name of module used for decorating output.  For
# example, if the value is NAME, a _pretty_NAME module must
# exist and be importable.  Otherwise pretty will fall back
# to plain, which is also the default value.
#
SHELLFU_PRETTY=${SHELLFU_PRETTY:-plain}

#
# Usage mode -- set to 'subcommand' to remove first dash
#
SHELLFU_PRETTY_USAGE=${SHELLFU_PRETTY_USAGE:-}


__shellfu_pretty__init() {
    #
    # Import proper submodule
    #
    if shellfu try_import "_pretty_${SHELLFU_PRETTY}";
    then
        shellfu import "_pretty_${SHELLFU_PRETTY}"
        return 0
    else
        warn "falling back to _pretty_plain"
        SHELLFU_PRETTY=plain
        shellfu import "_pretty_${SHELLFU_PRETTY}"
    fi
}


##--------------------##
## PRINTING FRONT-END ##
##--------------------##

debug() {
    #
    # You already know what it does
    #
    # BTW, following are equivalent:
    #
    #     debug "var1=$var1" "var2=$var2" "result=$result"
    #     debug -v var1 var2 result
    #     debug -v "var@ result"
    #
    $SHELLFU_DEBUG || return 0
    _pretty__echo "$@"
}


debug_pipe() {
    #
    # Debug the whole pipe.
    #
    grep . \
        | while IFS= read line;
          do
              debug "|$1: '$line'"
              echos "$line"
          done
}


die() {
    #
    # A fatal error
    #
    _pretty__echo -t
    _pretty__echo "$@"
    exit_error
}


echos() {
    #
    # Safer version of echo able to echo "-n"
    #
    # Traditional echo is broken in that it does not
    # distinguish between string to print and its own switches
    # (-e, -E or -n), leading to unexpected behavior.
    #
    # This echo version circumvents this by using printf.
    #
    printf -- '%s\n' "$@"
}


mkusage() {
    #
    # Echo out usage patterns and (by default) `exit 2`
    #
    #     mkusage [-e STATUS] [-E] pattern [patern...]
    #
    # Each pattern is prefixed by "usage: " and a resolved
    # script/function name to produce traditional usage hint.
    # By default, will exit with status *SHELLFU_EXIT_USAGE*, you
    # can use `-e STATUS` to specify other number. `-k` can be
    # used as shorthand for `-e 0` (e.g. if user asked for
    # the patterns)
    #
    # Use `-E` to prevent exiting; in that case, the status
    # that would be fed to `exit` will be used as exit status
    # of this function.
    #
    # Use "--" to delimit end of arguments processed by mkusage.
    #
    # Setting SHELLFU_PRETTY_USAGE to 'subcommand' has effect that
    # in each call, first hyphen in the script's "own" name is
    # replaced by space; this is used for git-like sub-commands,
    # where the actual binary name is e.g. "git-add" for
    # "git add".
    #
    # Recommended usage is to define usage() in your script and
    # use this in its body.  That way you only need to define
    # usage patterns once and skip to them from any place where
    # you detect incorrect usage.
    #
    local es=$SHELLFU_EXIT_USAGE
    local doexit=true
    while true; do case "$1" in
        -e) es="$2";            shift 2 || return 2 ;;
        -E) doexit=false;       shift        ;;
        -k) es=$SHELLFU_EXIT_OK; shift       ;;
        --)                     shift; break ;;
        *)                      break        ;;
    esac done
    _pretty__echo -u "$@";
    $doexit && exit "$es"
    return "$es"
}


mkhelp() {
    #
    # Echo out help text
    #
    #     mkhelp [-e STATUS] [-E] arg...
    #
    # By default, will exit with status *SHELLFU_EXIT_OK*, you
    # can use `-e STATUS` to specify other number.
    #
    # Use `-E` to prevent exiting; in that case, the status
    # that would be fed to `exit` will be used as exit status
    # of this function.
    #
    # Use "--" to delimit end of arguments processed by mkhelp
    #
    local es=$SHELLFU_EXIT_OK
    local doexit=true
    while true; do case "$1" in
        -e) es="$2";            shift 2 || return 2 ;;
        -E) doexit=false;       shift               ;;
        --)                     shift; break        ;;
        *)                      break               ;;
    esac done
    _pretty__echo "$@"
    $doexit && exit "$es"
    return "$es"
}


think() {
    #
    # If verbose is on, think loud
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    $SHELLFU_VERBOSE || return 0
    _pretty__echo "$@"
}


warn() {
    #
    # Warn them
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    _pretty__echo "$@"
}


##----------##
## BACK-END ##
##----------##


_pretty__cat() {
    #
    # `cat` but without starting a process
    #
    # Used to avoid spanning new process where stream handler is chosen
    # based on some logic
    #
    while IFS= read line;
    do echos "$line"; done
}


_pretty__get_caller() {
    #
    # Get first user function from stack
    #
    local fname
    for fname in "${FUNCNAME[@]}";
    do
        _pretty__is_internal && continue
        _pretty__is_frontend && continue
        test "$fname" = "usage" && continue
        echos "$fname()"
        return
    done
}

_pretty__get_frontend() {
    #
    # Get entry point function name from stack
    #
    local fname
    for fname in "${FUNCNAME[@]}";
    do
        _pretty__is_internal && continue
        _pretty__is_frontend && echos "$fname" && return 0
        echo "do not call _pretty_* directly: $fname" >&2
        return "$SHELLFU_EXIT_USAGE"
    done
}

_pretty__is_frontend() {
    #
    # True if $fname is one of our "frontends"
    #
    case "$fname" in
        debug)      return 0 ;;
        debug_pipe) return 0 ;;
        die)        return 0 ;;
        mkhelp)     return 0 ;;
        think)      return 0 ;;
        mkusage)    return 0 ;;
        warn)       return 0 ;;
    esac
    return 1
}

_pretty__is_internal() {
    #
    # True if $fname is our internal function
    #
    case "$fname" in
        _pretty__*) return 0 ;;
        *)          return 1 ;;
    esac
}

_pretty__echo() {
    #
    # A smarter echo backend
    #
    # A smarter backend for debug, warn, think, die and
    # mkusage.
    #
    #   -c cmd      echo output of a command
    #   -f file     echo output of a file (- for stdin)
    #   -l line [line...] echo each line separately
    #   -t          add stack trace to output
    #   -u patt [patt...] convert each patt to usage pattern
    #   -v var [var...] show contents of each var
    #
    local frontend  # who (of pretty.sh) was called (=> prettyprinter choice)
    local caller    # which user's function (or script) called it
    #     ^      ^    eg. if user calls 'debug hello' from function 'foo', then
    #                 frontend is 'debug' and caller is 'foo'.
    local provider  # which provider (_pretty__echo_*()) to use
    frontend="$(_pretty__get_frontend)" || exit_usage
    caller="$(_pretty__get_caller)"
    test "$caller"   == "main()"        && caller="$(basename "$0")"
    test "$caller"   == "usage()"       && frontend=mkusage
    test "$frontend" == "debug_pipe()"  && frontend=debug
    while true; do case $1 in
        -c|--cmd)   provider=cmd;   shift; break ;;
        -f|--files) provider=files; shift; break ;;
        -l|--lines) provider=lines; shift; break ;;
        -t|--trace) provider=trace; shift; break ;;
        -u|--usage) provider=usage; shift; break ;;
        -v|--vars)  provider=vars;  shift; break ;;
        *)          provider=args;         break ;;
    esac done
    _pretty__echo_$provider "$@" \
      | _pretty__$frontend >&2
}


_pretty__echo_args() {
    #
    # The simplest (but safe) printing of args
    #
    echos "$*"
}


_pretty__echo_cmd() {
    #
    # Print command line, launch it and report exit status
    #
    local c=\$
    test "$(id -u)" -eq 0 && c="#"
    echo "$c $*"
    "$@"
    echo "^ exit status: $?"
}


_pretty__echo_files() {
    #
    # Print names and contents of existing files
    #
    local fp
    for fp in "$@";
    do
        if test "$fp" = "-";
        then
            cat
        elif test -s "$fp" || test "$fp" = "/dev/stdin";
        then
            echo "-- $fp --"
            cat "$fp"
        fi
    done
}


_pretty__echo_lines() {
    #
    # Echo each argument as a separate line
    #
    local l;
    for l in "$@"; do _pretty__echo_args "$l"; done
}


_pretty__echo_trace() {
    #
    # Print "decorated" call trace (only in debug mode)
    #
    $SHELLFU_DEBUG || return 0
    local depth
    echo "== trace =="
    for depth in $(seq 0 ${#FUNCNAME});
    do
        caller "$depth" || break
    done \
      | tail -n +3 \
      | sed -e '
            s/^\([^ ]\+\) \([^ ]\+\) \(.*\)/\3:\1:\2()/
            #  ^line^,    ^func^,    ^file^
            1   s/^/ -> /g
            2,$ s/^/    /
        ' \
      | tac
}


_pretty__echo_help() {
    local oldverbose="$SHELLFU_VERBOSE"
    think -l "$@"
    SHELLFU_VERBOSE=$oldverbose
}

_pretty__echo_usage() {
    #
    # Print each argument as usage pattern
    #
    local self
    case "$SHELLFU_PRETTY_USAGE" in
        subcommand) self="${caller/-/ }" ;;
        *)          self="$caller" ;;
    esac
    local u
    for u in "$@";
    do echo "usage: $self $u"
    done
}


_pretty__echo_vars() {
    #
    # Report value of each named variable
    #
    # If the last char is '@', try to use it as wildcard
    #
    local varname
    for varname in "$@";
    do
        if ! _pretty__is_word "$varname";
        then
            warn "unsafe value skipped: $varname";
            continue
        fi
        _pretty__echo "$varname='$(echos "${!varname}")'"
    done
}


_pretty__is_word() {
    #
    # Check if $1 contains only alphanumeric chars or _
    #
    local tainted="$1"
    local clean=$(tr -c -d '_[:alnum:]' <<< "$tainted")
    test "$tainted" = "$clean"
}

==> v0.8.4.zip: shellfu/src/include/pretty.sh <==
#!/bin/bash

shellfu import exit

#
# Name of pretty-printer module
#
# Friendly name of module used for decorating output.  For
# example, if the value is NAME, a _pretty_NAME module must
# exist and be importable.  Otherwise pretty will fall back
# to plain, which is also the default value.
#
SHELLFU_PRETTY=${SHELLFU_PRETTY:-plain}

#
# Usage mode -- set to 'subcommand' to remove first dash
#
SHELLFU_PRETTY_USAGE=${SHELLFU_PRETTY_USAGE:-}


__shellfu_pretty__init() {
    #
    # Import proper submodule
    #
    if shellfu try_import "_pretty_${SHELLFU_PRETTY}";
    then
        shellfu import "_pretty_${SHELLFU_PRETTY}"
        return 0
    else
        warn "falling back to _pretty_plain"
        SHELLFU_PRETTY=plain
        shellfu import "_pretty_${SHELLFU_PRETTY}"
    fi
}


##--------------------##
## PRINTING FRONT-END ##
##--------------------##

debug() {
    #
    # You already know what it does
    #
    # BTW, following are equivalent:
    #
    #     debug "var1=$var1" "var2=$var2" "result=$result"
    #     debug -v var1 var2 result
    #
    $SHELLFU_DEBUG || return 0
    _pretty__echo "$@"
}


debug_pipe() {
    #
    # Debug the whole pipe.
    #
    grep . \
        | while IFS= read line;
          do
              debug "|$1: '$line'"
              echos "$line"
          done
}


die() {
    #
    # A fatal error
    #
    _pretty__echo -t
    _pretty__echo "$@"
    exit_error
}


echos() {
    #
    # Safer version of echo able to echo "-n"
    #
    # Traditional echo is broken in that it does not
    # distinguish between string to print and its own switches
    # (-e, -E or -n), leading to unexpected behavior.
    #
    # This echo version circumvents this by using printf.
    #
    printf -- '%s\n' "$@"
}


mkusage() {
    #
    # Echo out usage patterns and (by default) `exit 2`
    #
    #     mkusage [-e STATUS] [-E] pattern [patern...]
    #
    # Each pattern is prefixed by "usage: " and a resolved
    # script/function name to produce traditional usage hint.
    # By default, will exit with status *SHELLFU_EXIT_USAGE*, you
    # can use `-e STATUS` to specify other number. `-k` can be
    # used as shorthand for `-e 0` (e.g. if user asked for
    # the patterns)
    #
    # Use `-E` to prevent exiting; in that case, the status
    # that would be fed to `exit` will be used as exit status
    # of this function.
    #
    # Use "--" to delimit end of arguments processed by mkusage.
    #
    # Setting SHELLFU_PRETTY_USAGE to 'subcommand' has effect that
    # in each call, first hyphen in the script's "own" name is
    # replaced by space; this is used for git-like sub-commands,
    # where the actual binary name is e.g. "git-add" for
    # "git add".
    #
    # Recommended usage is to define usage() in your script and
    # use this in its body.  That way you only need to define
    # usage patterns once and skip to them from any place where
    # you detect incorrect usage.
    #
    local es=$SHELLFU_EXIT_USAGE
    local doexit=true
    while true; do case "$1" in
        -e) es="$2";            shift 2 || return 2 ;;
        -E) doexit=false;       shift        ;;
        -k) es=$SHELLFU_EXIT_OK; shift       ;;
        --)                     shift; break ;;
        *)                      break        ;;
    esac done
    _pretty__echo -u "$@";
    $doexit && exit "$es"
    return "$es"
}


mkhelp() {
    #
    # Echo out help text
    #
    #     mkhelp [-e STATUS] [-E] arg...
    #
    # By default, will exit with status *SHELLFU_EXIT_OK*, you
    # can use `-e STATUS` to specify other number.
    #
    # Use `-E` to prevent exiting; in that case, the status
    # that would be fed to `exit` will be used as exit status
    # of this function.
    #
    # Use "--" to delimit end of arguments processed by mkhelp
    #
    local es=$SHELLFU_EXIT_OK
    local doexit=true
    while true; do case "$1" in
        -e) es="$2";            shift 2 || return 2 ;;
        -E) doexit=false;       shift               ;;
        --)                     shift; break        ;;
        *)                      break               ;;
    esac done
    _pretty__echo "$@"
    $doexit && exit "$es"
    return "$es"
}


think() {
    #
    # If verbose is on, think loud
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    $SHELLFU_VERBOSE || return 0
    _pretty__echo "$@"
}


warn() {
    #
    # Warn them
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    _pretty__echo "$@"
}


##----------##
## BACK-END ##
##----------##


_pretty__cat() {
    #
    # `cat` but without starting a process
    #
    # Used to avoid spanning new process where stream handler is chosen
    # based on some logic
    #
    while IFS= read line;
    do echos "$line"; done
}


_pretty__get_caller() {
    #
    # Get first user function from stack
    #
    local fname
    for fname in "${FUNCNAME[@]}";
    do
        _pretty__is_internal && continue
        _pretty__is_frontend && continue
        test "$fname" = "usage" && continue
        echos "$fname()"
        return
    done
}

_pretty__get_frontend() {
    #
    # Get entry point function name from stack
    #
    local fname
    for fname in "${FUNCNAME[@]}";
    do
        _pretty__is_internal && continue
        _pretty__is_frontend && echos "$fname" && return 0
        echo "do not call _pretty_* directly: $fname" >&2
        return "$SHELLFU_EXIT_USAGE"
    done
}

_pretty__is_excluded() {
    #
    # True if $caller is excluded based on SHELLFU_DEBUG_EXCLUDE
    #
    # Check SHELLFU_DEBUG_EXCLUDE to see if $caller (using only module name
    # part, if possible) should be muted from debugging.
    #
    local listed        # item listed in SHELLFU_DEBUG_EXCLUDE
    local name          # module part of caller's name
    local qualified     # is caller "qualified" (ac. to shellfu scheme)?
    name=${caller%()}
    case "$name" in
        __*__*) qualified=true  ;;
        __*)    qualified=false ;;
        *__*)   qualified=true  ;;
        *)      qualified=false ;;
    esac
    if $qualified;
    then
        # we'll use only the module part of the name
        name=${name#_}      # drop one "internal" prefix
        name=${name#_}      # drop yet another one
        name=${name%__*}    # drop funcname
    fi
    for listed in ${SHELLFU_DEBUG_EXCLUDE//,/ };
    do
        test "$name" = "$listed" && return 0
    done
    return 1
}

_pretty__is_frontend() {
    #
    # True if $fname is one of our "frontends"
    #
    case "$fname" in
        debug)      return 0 ;;
        debug_pipe) return 0 ;;
        die)        return 0 ;;
        mkhelp)     return 0 ;;
        think)      return 0 ;;
        mkusage)    return 0 ;;
        warn)       return 0 ;;
    esac
    return 1
}

_pretty__is_internal() {
    #
    # True if $fname is our internal function
    #
    case "$fname" in
        _pretty__*) return 0 ;;
        *)          return 1 ;;
    esac
}

_pretty__echo() {
    #
    # A smarter echo backend
    #
    # A smarter backend for debug, warn, think, die and
    # mkusage.
    #
    #   -c cmd      echo output of a command
    #   -f file     echo output of a file (- for stdin)
    #   -l line [line...] echo each line separately
    #   -t          add stack trace to output
    #   -u patt [patt...] convert each patt to usage pattern
    #   -v var [var...] show contents of each var
    #
    local frontend  # who (of pretty.sh) was called (=> prettyprinter choice)
    local caller    # which user's function (or script) called it
    #     ^      ^    eg. if user calls 'debug hello' from function 'foo', then
    #     :      :..   *  frontend is 'debug'
    #     :.........   *  and caller is 'foo'.
    local provider  # which provider (_pretty__echo_*()) to use
    frontend="$(_pretty__get_frontend)" || exit_usage
    caller="$(_pretty__get_caller)"
    $SHELLFU_DEBUG && _pretty__is_excluded "$caller" && return 0
    test "$caller"   == "main()"        && caller="$(basename "$0")"
    test "$caller"   == "usage()"       && frontend=mkusage
    test "$frontend" == "debug_pipe()"  && frontend=debug
    while true; do case $1 in
        -c|--cmd)   provider=cmd;   shift; break ;;
        -f|--files) provider=files; shift; break ;;
        -l|--lines) provider=lines; shift; break ;;
        -t|--trace) provider=trace; shift; break ;;
        -u|--usage) provider=usage; shift; break ;;
        -v|--vars)  provider=vars;  shift; break ;;
        *)          provider=args;         break ;;
    esac done
    _pretty__echo_$provider "$@" \
      | _pretty__$frontend >&2
}


_pretty__echo_args() {
    #
    # The simplest (but safe) printing of args
    #
    echos "$*"
}


_pretty__echo_cmd() {
    #
    # Print command line, launch it and report exit status
    #
    local c=\$
    test "$(id -u)" -eq 0 && c="#"
    echo "$c $*"
    "$@"
    echo "^ exit status: $?"
}


_pretty__echo_files() {
    #
    # Print names and contents of existing files
    #
    local fp
    for fp in "$@";
    do
        if test "$fp" = "-";
        then
            cat
        elif test -s "$fp" || test "$fp" = "/dev/stdin";
        then
            echo "-- $fp --"
            cat "$fp"
        fi
    done
}


_pretty__echo_lines() {
    #
    # Echo each argument as a separate line
    #
    local l;
    for l in "$@"; do _pretty__echo_args "$l"; done
}


_pretty__echo_trace() {
    #
    # Print "decorated" call trace (only in debug mode)
    #
    $SHELLFU_DEBUG || return 0
    local depth
    echo "== trace =="
    for depth in $(seq 0 ${#FUNCNAME});
    do
        caller "$depth" || break
    done \
      | tail -n +3 \
      | sed -e '
            s/^\([^ ]\+\) \([^ ]\+\) \(.*\)/\3:\1:\2()/
            #  ^line^,    ^func^,    ^file^
            1   s/^/ -> /g
            2,$ s/^/    /
        ' \
      | tac
}


_pretty__echo_help() {
    local oldverbose="$SHELLFU_VERBOSE"
    think -l "$@"
    SHELLFU_VERBOSE=$oldverbose
}

_pretty__echo_usage() {
    #
    # Print each argument as usage pattern
    #
    local self
    case "$SHELLFU_PRETTY_USAGE" in
        subcommand) self="$(basename "$0")"; self="${self/-/ }" ;;
        *)          self="$caller" ;;
    esac
    local u
    for u in "$@";
    do echo "usage: $self $u"
    done
}


_pretty__echo_vars() {
    #
    # Report value of each named variable
    #
    local varname
    for varname in "$@";
    do
        if ! _pretty__is_word "$varname";
        then
            warn "unsafe value skipped: $varname";
            continue
        fi
        _pretty__echo "$varname='$(echos "${!varname}")'"
    done
}


_pretty__is_word() {
    #
    # Check if $1 contains only alphanumeric chars or _
    #
    local tainted="$1"
    local clean=$(tr -c -d '_[:alnum:]' <<< "$tainted")
    test "$tainted" = "$clean"
}
